
// =======================================================================
//
// RDR110.c
//
// PIC program for a Low Frequency Transponder Reader
// equiped with a Texas Instruments TMS3705A chip.
//
// Reports transponders over a RS232 connection to a PC using TI microreader
// protocol and shows information (Tag-ID, errors, etc) on LCD screen.
//
// Author:    Rob Hamerling
//
// (c) Copyright R. Hamerling 2004,2008. All rights reserved.
//
// =======================================================================
//
// DESCRIPTION
//
// General
//
// This code assumes that the transmitter/receiver functions are
// performed by a Texas Instruments TMS3705A chip.
// - incoming data from the SCIO pin (reversed polarity)
// - control of this chip via the TXCT pin
//
// =======================================================================
// Implementation
//
// Designed for Microchip PICmicro type 16F690, 16F689 (and 16F687).
// Other PICS could be used, but:
//  - the pin assignments in the picture below might have to be changed
//    to the appropriate ports.
//  - code may have to be adapted because of:
//     - port re-assignments
//     - bank switching optimisations
//     - disabling of unused PIC features which are enabled by default
//
// The RS232 interface chip may be any converter chip.
// A MAX202 or similar with small or no external capacitors is preferred.
// Note: The internal charge pump of the MAX202 might give interference
//       with the TMS3705A antenne circuit.
//
// The communications with the PC takes the only available UART.
// Communications with the TMS3705A chip uses a synchronous protocol.
//  - The Mode Control Register is written to inform the 3705 that
//    Synchronous protocol is used for reading (default is Asynchronous).
//  - Between characters the program waits for the start-bit level on
//    the SCIO-pin of the TMS3705.
//  - When startbit detected TXCT is toggled to collect 8 data bits and
//    1 stopbit.
//
// ---------------------------------------------------------------------
//
// Simplified schematics of reader
//
// This setup is known as "Prototype 3".
//
//                                                             COMx    DB9
//       TMS3705A          PIC16F690           MAX202          RS232   pin
//
//          +------------------+------------------+------ +5V
//          |                  |                  |
//     +----|----+      +------|------+     +-----|-----+
//     |   VDD   |      |     Vdd     |     |    Vcc    |
//     |         |      |             |     |           |
// ----|ANT1 SCIO|-->---|RB4       RB6|---->|T2in  T2out|--->-- CTS --- 8
// Ant.|         |      |             |     |           |
// -----ANT2 TXCT|--<---|RC5          |     |           |
//     |         |      |           RX|<----|R1out  R1in|---<-- TxD --- 3
//     |   VSS   |      |           TX|---->|T1in  T1out|--->-- RxD --- 2
//     +----|----+      |             |     |           |
//          |           |     Vss     |     |    Gnd    |
//          |           +------|------+     +-----|-----+
//          |                  |                  |
//          +------------------+------------------+------------ GND --- 5
//
//                                                .-----.
//                                             +--| 2K2 |-----< DTR --- 4
//                       Recommended wraps:    |  `-----'
//                                             +--------------> DSR --- 6
//                                             +--------------> DCD --- 1
//
// -------------------------------------------------------------------
// Some basic PIC and RS232 knowledge will be needed to fully understand
// the data flow and control signalling used in this program.
//
// The PIC ports use positive logic:
// '1' is positive voltage, '0' is ground.
//
// This program uses positive logic for boolean variables:
// the symbol TRUE for '1', the symbol FALSE for '0'.
// Signal levels: 0 = 0V,  1 = +5V
//
// In the RS232 standard:
// - Negative voltage ('mark') means OFF for control signals, and
//   indicates 1 (one) for data signals (start-, data-, stop-bits).
// - Positive voltage ('space') means ON for control signals and
//   0 (zero) for start-, data- and stop-bits.
// - Signal levels: 'mark' = -3V to -15V,  'space' = +3V to +15V
//
// The MAX232 is not only a voltage level CONverter (between TTL and RS232)
// but also a signal INverter. You should be aware of the following:
// - The inversion of PIC data-in and data-out by the MAX232 is required
//   to convert data-, start- and stop-bits to/from the corresponding
//   RS232 polarity. So nothing special has to be done in the program
//   because the USART of the PIC uses 'inverted' signal levels!
// - For RTS and CTS the inversion by the MAX232 is NOT desired.
//   Therefore the program has to use inverted signalling for RTS and CTS:
//   'FALSE' is used for ON and 'TRUE' for OFF with RTS/CTS signals!
//   As a reminder for this 'inverted' logic the signals are called
//   here CTSinv and RTSinv (e.g. CTSinv=true, means CTS is false!).
//   (Note: RTS not used anymore since version 0.4.5)
//
// Remember also: RxD of RS232-plug connects to TX of PIC
//                TxD of RS232-plug connects to RX of PIC
// ---------------------------------------------------------------------

// PIC type specified on commandline (in make file)

// Note: When a bootloader is used, the following statements
//       are required (without bootlader they won't harm!!).

#pragma  cdata[0] = 0x0180 + 0x0A               // CLRF PCLATH
#pragma  cdata[1] = 0                           // NOP
#pragma  cdata[2] = 0                           // NOP

#pragma  resetVector 3                          // position of 'goto main'

// -----------------------------------------------------------------

// This code contains some compile-time variables (macros) to
// control which [optional] parts of the code should be included.
// Because of memory constraints probably not all options can be included
// at the same time!
//
// By defining the linesize of the used LCD screen code for it
// will be generated at compile time.
// The LCD will show the hex representation of the last received key.
// The LCD is supposed to be controlled by a standard HD44780
// controller (or a compatible) chip. The schematics above don't
// show the LCD connection, see for this the complete schematics.
// A 'standard' 16-pin header is used. With 4-bit data transfer and delays
// (no busy-reads) only 6 pins of the PIC are needed to control the
// LCD screen. Power and backlight are permanently connected.

#define LCD_LINESIZE 16

// NOTE: The code assumes a 2x16 display, so the value '16' is mandatory.


// ------------------------------------------------------------------
// The reader uses S2000 Microreader protocol.
// There are 2 alternatives: Loconet or S88 protocol.
//
// By defining the macro 'LOCONET' Loconet protocol is selected.
// #define LOCONET
// NOTE: ==>> Loconet protocol is not (yet) provided with this source,
//            and requires a different hardware interface.
//
// By defining the macro 'S88' S88 feedback protocol is selected.
// #define S88
// NOTE: ==>> S88 protocol is not (yet) provided by this source,
//            and requires a different hardware interface.

// -----------------------------------------------------------------

#pragma  config  &= ~0b11.1111.1111.1111        // all config bits OFF

#if  defined(_16F690) | defined(_16F689) | defined(_16F687)

  #pragma  config  |=  0b11.0011.1110.0010
//                                     xxx        Fosc HS (16 MHz crystal)
//                                    x           WDT disabled
//                                  x             PWRT enabled
//                                 x              /MCLR is /MCLR
//                                x               no code memory protection
//                               x                no data memory protection
//                            xx                  BOR enabled, SBOREN disabled
//                           x                    Int/Ext switchover disabled
//                          x                     Failsafe Clock mon. disabled
//                       xx                       fixed 1
//

#else

  #error Program does not support the specified PIC (chip)!

#endif

#include "Rdr110.h"                             // program specific defines

// ID bytes:

#pragma  config ID[0] = 0                       // )
#pragma  config ID[1] = VERSIONMAJOR            // ) firmware ..
#pragma  config ID[2] = VERSIONMINOR            // ) .. version
#pragma  config ID[3] = VERSIONSUFFIX           // )

// EEPROM data: not used by program, only for program identification

#pragma  cdata[0x2100]
#pragma  cdata.ID0    = 0
#pragma  cdata.ID1    = VERSIONMAJOR
#pragma  cdata.ID2    = VERSIONMINOR
#pragma  cdata.ID3    = VERSIONSUFFIX

#pragma  cdata.major  =  (__CC5X__ / 1000)
#pragma  cdata.minor  = ((__CC5X__ % 1000) / 100)
#pragma  cdata.minor2 = ((__CC5X__ % 100)  / 10)
#pragma  cdata.suffix =  (__CC5X__ % 10)

#pragma packedCdataStrings  0
#pragma  cdata.date   =  __DATE2__
#pragma packedCdataStrings  1


// ---------------------------------------------------------------------
// The receive and transmit buffers are explicitly located in
// RAM bank1 and bank2. All other variables are located in Bank 0.
// RAM bank switching can almost be eliminated, and unproductive bank
// switching is disabled by "#pragma update_RP 0" in most functions.
// Functions must return with RP0 and RP1 reset (0) to prevent errors,
// which can be ensured by "#pragma updateBank exit = 0". But this
// is in fact only required in functions which address registers in
// other banks directly, like picSetup().
// Note: The variables 'rcvbuf' and 'xmtbuf' are referenced only
// indirectly, which means effectively that RP0 and RP1 are irrelevant.
// ---------------------------------------------------------------------

// Receive and transmit buffers for communications with PC via UART

static bank1 uns8  xmtbuf[XMTBUFSIZE];          // circular output buffer
static bank2 uns8  rcvbuf[RCVBUFSIZE];          // circular input buffer

// Offsets in circular buffers for communications with PC via UART

static  uns8       OffsetXmtHead;               // offset next byte from appl
static  uns8       OffsetXmtTail;               // offset next byte to port
static  uns8       OffsetRcvHead;               // offset next byte from port
static  uns8       OffsetRcvTail;               // offset next byte to appl

// Receive buffer for frames from TMS3705A

static  union {SCIOFRAME  f;                    // frame structure
               uns8       a[sizeof(SCIOFRAME)]; // array
                 } ScioBuf;                     // buffer

// Other globals (mainly communications between ISR and main)

static  uns8     TMR0counter;                   // count TMR0 interrupts
static  uns16    CRC;                           // current CRC (calculated)
static  uns16    lastCRC;                       // last received CRC

#include <int16cxx.h>                           // interrupt support

// =======================================================================
// Interrupt service routine for:
// - TX       - UART data transmission (host)
// - RC       - UART data reception (host)
// - Timer 0  - 1 millisecond clock tick for charge/read/idle phase control
// =======================================================================

#pragma  origin  INTERRUPT_VECTOR               // required positioning

interrupt InterruptHandler(void) {

  int8  x;                                      // intermediate byte value
  uns8  saveFSR;                                // save/restore FSR

  int_save_registers
  saveFSR = FSR;                                // save FSR

#pragma update_RP 0

  if (RCIF == TRUE) {                           // UART receive interrupt
    if (OERR == TRUE  ||  FERR == TRUE) {       // framing/overrun errot
      x = RCREG;                                // flush hardware buffer
      while (OERR == TRUE) {                    // clear overrun state
        CREN = FALSE;                           // disable UART
        CREN = TRUE;                            // re-enable UART
        x = RCREG;                              // \  flush hardware buffers
        x = RCREG;                              // /
        };                                      // until no more overrun
      OffsetRcvTail = 0;                        // )
      OffsetRcvHead = 0;                        // ) flush circular buffer
      CTSinv = FALSE;                           // ensure CTS true
      }
    else {                                      // data without errors
      rcvbuf[OffsetRcvHead] = RCREG;            // move byte to rcv buffer
      x = (OffsetRcvHead + 1) & (RCVBUFSIZE - 1);   // update offset (wrap)
      if (x != OffsetRcvTail)                   // buffer not yet full
        OffsetRcvHead = x;                      // update offset
                                                // (else discard byte, CTS
                                                //  flow control has failed)
      if (CTSinv == FALSE) {                    // CTS is TRUE
        x = OffsetRcvTail - OffsetRcvHead;      // offset difference
        if (x < 0)                              // wrapping effect
          x += RCVBUFSIZE;                      // wrapping correction
        if (x < DELTA)                          // circular buffer almost full
          CTSinv = TRUE;                        // set CTS FALSE
        }
      }
    }

#pragma update_RP 1

  else if ((TXIF == TRUE) && (TXIE == TRUE)) {  // UART transmit interrupt
    if (OffsetXmtTail != OffsetXmtHead) {       // still data in xmit buffer
      x = xmtbuf[OffsetXmtTail];                // fetch next byte to xmit
      if (++OffsetXmtTail >= XMTBUFSIZE)        // beyond buffer boundary
        OffsetXmtTail = 0;                      // wrap to begin
      if (OffsetXmtTail == OffsetXmtHead)       // is last byte to be xmt'd
        TXIE = FALSE;                           // disable xmit interrupts
      TXREG = x;                                // actually xmit char
      }
    }

  else if (T0IF == TRUE) {                      // TMR0 interrupt
    TMR0 += DELAY_TMR0_1MS_LOAD_VALUE;          // reload timer
    ++TMR0counter;                              // count 1 ms periods
    T0IF = FALSE;                               // clear interrupt flag
    }

  FSR = saveFSR;                                // restore FSR
  int_restore_registers

  }


// =======================================================================
//
//     M A I N L I N E
//
// The PIC is initialized and put into the required mode.
// The software then goes into an infinite loop.
// A cycle consists of the following phases:
//
// - charge-phase
//   + TXCT is false (low voltage)
//   + diagnostic byte will be received
//
// - read-phase
//   + TXCT is true (high voltage)
//   + Read and green LEDs are swiched off
//   + Collecting data from TMS3705A chip
//   + When no data is received within 3 ms the read-phase is aborted
//     (normal duration 20 ms)
//
// - idle-phase
//   + When no complete and valid frame is received a 'no-read'
//     will be reported to the host (once) and the charge-phase is
//     re-entered immediately
//   + When a valid frame received the key will be reported to host (once),
//     and shown on the LCD screen. Th green LED is switched on.
//   + When an incomplete or invalid frame is received the red LED will
//     be switched on.
//   + Any incoming command from PC is handled.
//
// - (end of idle-phase)
//   + cycle is restarted
//
// Only transponder changes will be reported to the host and LCD.
// When the same transponder is read consecutively only the first time
// is reported. As soon as no frame is received a 'no-read' is reported.
// When a good frame is received again after a 'no-read' it will be
// reported, even if it is the same as the previous good frame.
//
// NOTE: The check if the same tag is received is currently done by
//       comparing the (correct) CRC values. This is not full proof,
//       different keys may have the same CRC!
//
// When a good frame is received with an indication R/O or R/W transponder,
// a 'brief' report is send to the computer, otherwise a 'full' report is
// sent (with 'other' in the status byte).
//
// To be decided:
// Bad frames are currently considered to be 'boundary' effects, probably
// a transponder is entering or leaving the antenna field of the reader.
// The error-LED signals a bad frame, and when the previous read was a
// complete frame a 'no-read' is reported.
//
// =======================================================================
extern void main(void) {

  bit     bFrameHandled;                        // last received frame handled
  bit     bNoTagReported;                       // no-tag (/error) reported
  uns8    k;                                    // bytes received from TMS3705
  uns8    x;                                    // byte input buffer

  picSetup();                                   // PIC setup

#pragma update_RP 0                             // bank switching off

#if defined(LCD_LINESIZE)
  lcdSetup();                                   // init screen
#endif

  OffsetXmtTail = 0;                            // (re-)init ..
  OffsetXmtHead = 0;                            //  .. input and ..
  OffsetRcvHead = 0;                            //   .. output ..
  OffsetRcvTail = 0;                            //    .. buffer offsets
  CTSinv = FALSE;                               // accept data from host

  TMR0counter = TIMEOUT_IDLE;                   // start at end of cycle!
  bNoTagReported = FALSE;                       // force 'no' if no tag
  lastCRC = 0xFFFF;                             // force report ..
                                                // .. if tag present at startup

  for (;;) {                                    // loop forever

    if (TMR0counter < TIMEOUT_CHARGE) {  /* -------- charge phase --------- */
      TXCT = FALSE;                             // TMS3705A in charge phase
      if (SCIO == TRUE) {                       // 'diagnostic' byte available
        TXCT = TRUE;                            // (temp) reset charge level
        delay100microsecond(1);                 // wait a little
        x = receiveByte();                      // get byte
        ScioBuf.a[0] = x;                       // put in frame buffer
        k = 1;                                  // only 1 byte expected
        TXCT = FALSE;                           // restore charge level
        }
      // Charge phase might be abandoned when no 'normal' diagnostic received.
      }

    else if (TMR0counter < TIMEOUT_READ) { /* ------ read phase ----------- */
      TXCT = TRUE;                              // end charging
      READY_LED = FALSE;                        // reset tag key read signal
      ERROR_LED = FALSE;                        // reset BCC error signal
      bFrameHandled = FALSE;                    // new frame not handled yet
      if (SCIO == TRUE) {                       // byte available
        x = receiveByte();                      // get byte (synchr.)
        if (k < sizeof(ScioBuf)) {              // discard overflow bytes
          ScioBuf.a[k++] = x;                   // copy to frame buffer
          }
        }
                                                // check 'no-read' condition
      if (TMR0counter > (TIMEOUT_CHARGE + 3)  &&  // some ms in read phase
          k < 2)                                // still no data received
        TMR0counter = TIMEOUT_READ;             // abort read-phase
      }

    else if (TMR0counter < TIMEOUT_IDLE) { /* --------- idle phase -------- */
      if ( bFrameHandled == FALSE) {            // last rcv'd frame not handled
        bFrameHandled = TRUE;                   // prevent re-processing
        calcCRC();                              // calculate CRC
        if ( k >= sizeof(SCIOFRAME)  &&         // complete frame
             ScioBuf.f.Start == ScioBuf.f.Stop  &&   // startbyte == stopbyte
            (ScioBuf.f.Start == START_RO || ScioBuf.f.Start == START_RW) &&
             CRC == ScioBuf.f.BCC ) {           // calc'd  BCC == rcv'd BCC
          READY_LED = TRUE;                     // signal 'tag correct'
          bNoTagReported = FALSE;               // reset no-read flag
          if (CRC != lastCRC) {                 // (too?) simple check new tag
            lastCRC = CRC;                      // remember new CRC
            hostSendReport(RPT_STS);            // report tag to host
#if defined(LCD_LINESIZE)
            lcdShowKey();                       // show tag on LCD screen
#endif
            }
          }
        else {                                  // incomplete frame or error
          if (bNoTagReported == FALSE) {        // not reported 'no-read' yet
            if (ScioBuf.f.Diag == 0x50) {       // normal diagnostic
              bNoTagReported = TRUE;            // reported now!
              hostSendReport(RPT_NR);           // report 'no-read' to host
              }
            else                                // error condition
              ERROR_LED = TRUE;                 // signal error
#if defined(LCD_LINESIZE)
            lcdShowDiag(ScioBuf.f.Diag);        // show diagnostic byte
#endif
            lastCRC += CCITTPOLY;               // 'forget' last key
            }
          TMR0counter = TIMEOUT_IDLE;           // terminate idle-phase
          }
        }
      hostExecCommand();                        // exec host command (if any)
      }

    else {                                /* ---- end of cycle --------- */
      transferMCR(0b0011.1110);                 // start charge with write MCR
      k = 0;                                    // new frame expected
      TMR0counter = 0;                          // back to charge-phase
      }

    } /* forever */

  }

#pragma update_RP 1                             // bank switching on


// -----------------------------------------------------------------
// Calculate CRC of received key
//
// Store result in global variable 'CRC'
//
// Remarks:
// CRC is calculated from ScioBuf.f.Key byte-by-byte left-to-right.
// (which is in sequence of reception from TMS3705A).
// The CRC calculation is according to the CCITT method.
//
// -----------------------------------------------------------------

#pragma update_RP 0                             // bank switching off

static  void  calcCRC(void) {

  uns8   i, j;                                  // counter
  uns8   x;                                     // intermediate
  uns16  tmp;

  CRC = 0x0000;                                 // init calculated CRC
  for (i = 0; i < 8; i++) {                     // whole key
    x = ScioBuf.f.Key[i];                       // byte of TagKey
    j = 8;                                      // 8 bits per byte
    do {                                        // all bits in byte
      tmp = (uns16)x << 15;                     // shift to MSbit
      tmp = CRC ^ tmp;                          //
      CRC <<= 1;                                // shift
      if (tmp & 0x8000)                         // MSbit 1
        CRC ^= CCITTPOLY;                       // apply mask
      x >>= 1;                                  // next higher order bit
      } while (--j > 0);                        // all bits of byte
    };                                          // all bytes of key

  tmp = CRC;                                    // take a copy
  i = 16;                                       // 16 bits in a word
  do {                                          // reverse bit sequence
    tmp = rr(tmp);                              // LSb(source) -> carry
    CRC = rl(CRC);                              // Carry -> LSb(destination)
    } while (--i > 0);                          // all bits

  return;                                       // result is stored in CRC
  }

#pragma update_RP 1                             // bank switching on


// ----------------------------------------------------------------
//  Return a received command from PC in the circular input buffer
//
//    input: - pointer to caller's (lineair) buffer
//
//   output: - when valid command received it will be executed
//
//   returns nothing
//
// remarks: - returns 0 when not a complete command present
//          - rise CTS when receive buffer has more than <DELTA>
//            bytes free space after delivering data to caller.
//
//
//  Layout of reader commands
//
//  |<------------------ frame ---------------------->|
//  |                                                 |
//  +-----+--------+---------+------------------+-----+
//  | SOH | LENGTH | COMMAND |       DATA       | BCC |
//  +-----+--------+---------+------------------+-----+
//        |   .    |                            |
//        |   .....|.............               |
//        |        |            .               |
//        |        |            V               |
//        |        |<-------------------------->|
//        |                                     |
//        |<------------- packet -------------->|
//
//
//  'Frame' is the unit of data transferred over an RS232 interface
//  'Packet' is the part of the frame passed to the application
//
// Only when a frame is completely received and the BCC is OK
// the command will be executed.
// The packet buffer is supposed to be large enough to accept
// the largest supported packet.
//
// ----------------------------------------------------------------

#pragma update_RP = 0                           // bank switching off

static  void  hostExecCommand(void) {

  uns8  i;                                      // counter(s)
  int8  x;                                      // received character
  int8  BytesBuffered;                          // bytes in receive buffer
  uns8  PacketLength;                           // bytes to user
  uns8  ScanOffset;                             // bytes in receive buffer
  uns8  XOR;                                    // calculated BCC
  uns8  buffer[CMDBUFSIZE];                     // host command buffer

  for (;;) {                                    // 'break' to terminate loop

    PacketLength = 0;                           // default return: no command

    if (OffsetRcvTail == OffsetRcvHead)         // receive buffer empty
      break;                                    // exit from outer loop

    do {                                        // search start of command
      x = rcvbuf[OffsetRcvTail];                // next char
      if (x == SOH)                             // start of command
        break;                                  // yes, end search for it
      OffsetRcvTail = (OffsetRcvTail + 1) & (RCVBUFSIZE - 1);  // skip char
      } while (OffsetRcvTail != OffsetRcvHead);  // more data buffered

    if (x != SOH)                               // no start of command found
                                                // before buffer emtpy
      break;                                    // exit from outer loop
                                                // (PacketLength = 0)

                                                // keep 'OffsetRcvTail'
                                                // asis for possible rescan
    ScanOffset = (OffsetRcvTail + 1) & (RCVBUFSIZE - 1);  // shadow offset
    XOR = rcvbuf[ScanOffset];                   // get length field
                                                // and init BCC calculation
    BytesBuffered = OffsetRcvHead - ScanOffset; // offset difference
    if (BytesBuffered < 0)                      // wrapped
      BytesBuffered += RCVBUFSIZE;              // correction for wrapping
    if (XOR + 2 > BytesBuffered)                // frame incomplete?
      break;                                    // if yes, exit from outer
                                                // loop (PacketLength = 0)

    buffer[0] = XOR;                            // store length field
    PacketLength = XOR + 1;                     // # bytes to return
                                                // length + data (exlc BCC)
    for (i = 1; i <= PacketLength; i++) {       // data (incl BCC for XOR!)
      ScanOffset = (ScanOffset + 1) & (RCVBUFSIZE - 1);  // to next byte
      x = rcvbuf[ScanOffset];                   // fetch next byte ..
      buffer[i] = x;                            // .. copy to user buffer
      XOR ^= x;                                 // 'add' to calculated BCC
      }

    if (XOR == 0x00) {                          // BCC correct, frame OK!
      OffsetRcvTail = (ScanOffset + 1) & (RCVBUFSIZE - 1); // now remove frame!
      break;                                    // exit from outer loop
      }
                                                // incorrect frame received,
                                                // must rescan, beginning
                                                // after last detected SOH
    OffsetRcvTail = (OffsetRcvTail + 1) & (RCVBUFSIZE - 1);  // skip last SOH

    }

  // -- perform receive flow control --

  if (CTSinv == TRUE) {                         // CTS is FALSE
    x = OffsetRcvTail - OffsetRcvHead;          // offset difference
    if (x < 0)                                  // wrapping effect
      x += RCVBUFSIZE;                          // wrapping correction
    if (x >= DELTA)                             // enough free space now
      CTSinv = FALSE;                           // make CTS TRUE
    }

  // -- now exec the host command (if any recognised) --

  if (PacketLength > 0) {                       // complete command received

    if (buffer[1] == (COMMAND_POWER_BURST1 |    // reprogram transponder
                      COMMAND_POWER_BURST2 |
                      COMMAND_DATA))
      transferNewKey(buffer);                   // send new key to transponder

    else if ((buffer[1] & (COMMAND_NORMAL_MODE | COMMAND_LINE_MODE)) ==
                     COMMAND_SINGLE)            // single command
      hostSendReport(RPT_STS);                  // send status report

    else if ((buffer[1] & (COMMAND_NORMAL_MODE | COMMAND_LINE_MODE)) ==
                     COMMAND_VERSION) {         // query firmware version
      lastCRC += CCITTPOLY;                     // forget last key
      hostSendReport(RPT_SW);                   // report firmware version
      }

    else                                        // no other types sup'd yet
      hostSendReport(RPT_FULL);                 // send full report

    }

  return;                                       // # of bytes returned

  }

#pragma update_RP 1                             // bank switching on


// --------------------------------------------------------------
//  Put a single byte in circular transmit buffer for UART
//
//  input:   byte from caller
//
//  returns: transmitted byte
//
//  notes: - initiates transmission (interrupt handler)
//           when not currently transmitting
//         - spin when transmission buffer full
//           (wait for free buffer space)
//
// -----------------------------------------------
#pragma update_RP 0

static  uns8 hostSendByte(uns8  ucByte) {

  uns8  newOffset;                              // circular buffer pointer

  xmtbuf[OffsetXmtHead] = ucByte;               // put byte in buffer
  newOffset = OffsetXmtHead + 1;                // offset next char
  if (newOffset >= XMTBUFSIZE)                  // beyond buffer boundary
    newOffset = 0;                              // wrap to begin
  while (newOffset == OffsetXmtTail)            // buffer full!
    ;                                           // spin until something xmit'd
  OffsetXmtHead = newOffset;                    // update offset now
  RP0 = 1;                                      // TXIE is in bank1
  TXIE = TRUE;                                  // (re-)enable xmit interrupts
  RP0 = 0;                                      // return to bank 0
  return ucByte;                                // transmitted byte
  }

#pragma update_RP 1


// -----------------------------------------------------------------
// Send report to host
//
//  input:  report type
//
//  output: report will be inserted in RS232 output buffer
//
//  returns nothing
//
//  Formally description of micro-reader to PC protocol is in Texas
//  Instruments manual:
//  "Series 2000 Reader System RFM Sequence Control Reference Manual",
//  ("rfm_seq_control.pdf").
//
//  This reader does not support Multipage transponders, therefore
//  the statusbyte only reports:
//  - type of transponder (readonly or readwrite)
//  - validity of DBCC (received value must match calculated value)
//
// -----------------------------------------------------------------

#pragma update_RP 0                             // bank switching off

static  void  hostSendReport(uns8 type) {

  uns8   i;                                     // intermediate byte value
  uns8   XOR;                                   // accumulate BCC
  uns8   StatusByte;                            // status flags


  hostSendByte(SOH);                            // start of frame

  if (type == RPT_SW) {                         // report firmware version
    XOR  = hostSendByte(2);                     // length (fixed)
    XOR ^= hostSendByte(STATUS_TYPE_OTHER | STATUS_VERSION);  // Firmware
    XOR ^= hostSendByte((VERSIONMAJOR << 4) | VERSIONMINOR);  // Major.Minor
    }

  else if (type == RPT_NR) {                    // no read report
    XOR  = hostSendByte(1);                     // length (fixed)
    XOR ^= hostSendByte(STATUS_TYPE_OTHER);     // other, no startbyte, etc
    }

  else {                                        // transponder reports
    StatusByte = 0x00;                          // init
    if (ScioBuf.f.Start == START_RO)            // RO transponder
      StatusByte = STATUS_START_BYTE;           // only startbyte flag
    else if (ScioBuf.f.Start == START_RW)       // R/W transponder
      StatusByte = STATUS_START_BYTE | STATUS_TYPE_RW;   // start + R/W flag
    else                                        // not RO or RW transponder
      type = RPT_FULL;                          // force full report

    if (CRC == ScioBuf.f.BCC)                   // calculated matches received
      StatusByte |= STATUS_DBCC_OK;             // DBCC OK flag
    else                                        // CRC error
      type = RPT_FULL;                          // force full report

    if (type == RPT_STS) {                      // report status
      XOR  = hostSendByte(9);                   // length (fixed)
      XOR ^= hostSendByte(StatusByte);          // status byte (see above)
      for (i = 0; i < 8; ++i)                   // whole key
        XOR ^= hostSendByte(ScioBuf.f.Key[i]);  // LSB first
      }

    else {                                      // other (full)
      XOR  = hostSendByte(15);                  // length (fixed)
      XOR ^= hostSendByte(StatusByte);          // status byte (see above)
      for (i = 1; i < 15; i++)                  // the whole bunch:
        XOR ^= hostSendByte(ScioBuf.a[i]);      // start,key,CRC,stop,end
      }

    }

  hostSendByte(XOR);                            // end of frame (BCC)

  }

#pragma update_RP 1                             // bank switching on


// -----------------------------------------------------------------
// Receive byte from 3705 Synchronously
//
// input: nothing
//
// output: nothing
//
// returns received byte
//
// Note: This function is supposed to invokled only when SCIO has
//       become TRUE during charge or read phase.
// -----------------------------------------------------------------

#pragma update_RP 0                             // bank switching off

static  uns8  receiveByte(void) {

  uns8   i;                                     // counters
  uns8   x;                                     // input buffer

  for (i = 0; i < 9; i++) {                     // 8 bits + stopbit
    TXCT = FALSE;                               // falling edge of clock
    delay16microsecond(2);                      // half clock cycle
    if (i < 8) {                                // 8 data bits (skip stopbit)
      Carry = !SCIO;                            // 3705 uses inverted polarity
      x = rr(x);                                // shift into byte-buffer
      }
    TXCT = TRUE;                                // rising edge of clock
    delay16microsecond(2);                      // half clock cycle
    }
  return x;
  }

#pragma update_RP 1                             // bank switching on


// --------------------------------------------------------------
//  Transfer Mode control register
//
//  input:   MCR contents (1 byte)
//
//  returns: nothing
//
// -----------------------------------------------

static void transferMCR(uns8  mcr) {

#pragma update_RP 0

  uns8   i;                                     // counter(s)

  i = 8;                                        // 8 bits in a byte
  do {                                          // xmit in NRZ mode
    if (mcr & 0b0000.0001)                      // Least Significant bit first
      TXCT = TRUE;                              // 1-bit
    else
      TXCT = FALSE;                             // 0-bit
    delay16microsecond(8);                      // 128 s
    mcr >>= 1;                                  // next bit
    } while(--i > 0);                           // whole byte

  return;

  }

#pragma update_RP 1                             // bank switching on


// --------------------------------------------------------------
//  Transfer new key to TMS3705A
//
//  input:   pointer to packet from PC (past SOH)
//
//  returns: nothing
//
//  remarks: - This function asummes it is invoked in the IDLE
//             phase of the read process (see mainline).
//           - The contents of the buffer are assumed to contain
//             complete and information to transmit directly.
//
// -----------------------------------------------

#pragma update_RP 0

static  void  transferNewKey(uns8 *buf) {

  uns8  i, j;                                   // counters (bit/byte)
  uns8  x;                                      // current byte to xmit

  TXCT = FALSE;                                 // wakeup pulse (from sleep)
  delay16microsecond(3);                        // 50 s wakup pulse
  TXCT = TRUE;                                  // make sure TXCT is high
  delay100microsecond(20);                      // t_init (req'd 2.05 ms)

  transferMCR(0b0011.1110);                     // write MCR: default freq.

  TXCT = FALSE;                                 // set TXCT low
  i = 10;                                       // (100 s -> 1 ms)
  do {
    delay100microsecond(buf[2]);                // power burst 1
    } while (--i > 0);

  for (i = 5; i < 19; i++) {                    // 14 bytes in total
    x = buf[i];                                 // byte to trasnfer
    j = 8;                                      // 8 bits in a byte
    do {
      TXCT = TRUE;                              // pulse
      if (x & 0b0000.0001) {                    // '1' (high) bit
        delay100microsecond(toffH);             // pulse duration
        TXCT = FALSE;                           //
        delay100microsecond(tonH);              // pause duration
        }
      else {                                    // '0' (low) bit
        delay100microsecond(toffL);             // pulse duration
        TXCT = FALSE;                           //
        delay100microsecond(tonL);              // pause duration
        }
      x >>= 1;                                  // next bit
      } while (--j > 0);                        // all bits
    }                                           // all bytes

  TXCT = FALSE;                                 // set TXCT low
  i = 10;                                       // (100 s -> 1 ms)
  do {
    delay100microsecond(buf[3]);                // power burst 2
    } while (--i > 0);

  TMR0counter = TIMEOUT_CHARGE;                 // end charge fase,
                                                // start read phase in mainline
  return;
  }

#pragma update_RP 1                             // bank switching on

// -------------------------------------------------------------------
//  Delay by instruction loop
//  Inner loop: 14 s
//  Outer loop:  1 s
//  Call/return  1 s
// -------------------------------------------------------------------

#pragma update_RP 0

static void delay16microsecond(uns8  units) {

  do  {
    delay1mus(14);                              // 14 s
    } while (--units > 0);                      // number of units
  }

#pragma update_RP 1


// -------------------------------------------------------------------
//  Delay by instruction loop
//  Inner loop: 98 s
//  Outer loop:  1 s
//  Call/return  1 s
// -------------------------------------------------------------------

#pragma update_RP 0

static void delay100microsecond(uns8  units) {

  do  {
    delay1mus(98);                              // 98 s delay
    } while (--units > 0);                      // number of units
  }

#pragma update_RP 1


// =======================================================================
// Function to setup the general operational parameters of the chip
//
// Instructions are ordered for minimum bank switching.
//
// =======================================================================
static  void  picSetup(void) {

#pragma updateBank exit = 0                     // return in bank 0

  PORTA    = 0b0000.0000;                       // )
  PORTB    = 0b0000.0000;                       // ) all down
  PORTC    = 0b0000.0000;                       // )

  ADCON0   = 0b0000.0000;                       // disable ADC (digital I/O)
  T1CON    = 0b0000.0000;                       // Timer1 not used
#if defined(_16F690)
  CCP1CON  = 0b0000.0000;                       // Capt/Comp/PWM off
  T2CON    = 0b0000.0000;                       // Timer2 not used
#endif

  PIR1     = 0b0000.0000;                       // reset all ext. int. flags
  RCSTA    = 0b1001.0000;                       // SPEN,CREN on, rest off
  //           x . . . . . . SPEN               // serial port enable
  //              x  . . . . CREN               // continuous receive enable

  CM1CON0  = 0b0000.0000;                       // Comparator C1 off
  CM2CON0  = 0b0000.0000;                       // Comparator C2 off
  ANSEL    = 0b0000.0000;                       // ) all ports ..
  ANSELH   = 0b0000.0000;                       // ) .. digital I/O

  WPUB     = 0b0000.0000;                       // weak pullups port B
  IOCB     = 0b0000.0000;                       // no interrupt on change

  IOCA     = 0b0000.0000;                       // Interrupt on change (none)
  TRISA    = 0b0000.1000;                       // Input
  //                x  . . . /MCLR              // master reset

  OPTION   = 0b0000.0000 | DELAY_TMR0_1MS_PRE;  // prescaler for Timer0
  //           x . . . . . . RABPU              // no global weak pullup
  //            x  . . . . . INTEDG             // interrupt on falling edge
  //             x . . . . . T0CS               // clock source internal
  //              x  . . . . T0SE               // increment on low to high

  WPUA     = 0b0000.0000;                       // weak pullups port A

  TRISB    = 0b1011.0000;                       // Input pins: RX,TX, RB4
  //           x . . . . . . TX                 // transmit data UART (!)
  //             x   . . . . RX                 // receive data UART
  //              x  . . . . RB4                // receive data TMS3705A

  TRISC    = 0b0000.0000;                       // All output

  SPBRG    = BRGLOADCOUNT;                      // 8-bit baudrate clock divisor

  TXSTA    = 0b0010.0100;                       // TXEN,BRGH on; rest off
  //             x . . . . . . TXEN             // UART transmit int. enable
  //              x  . . . . . SYNC             // async mode
  //                 x . . . . BRGH             // high baud rate

  BAUDCTL  = 0b0000.0000;                       // no autobaud detect
                                                // 8 bit Baudrate generator
                                                // non inverted data

  PIE1     = 0b0010.0000;                       // external interrupts
  //             x . . . . . . RCIE             // UART receive int. enable
  //                                            // (xmit int. dynamically)

  INTCON   = 0b1110.0000;                       // interrupts
  //           x . . . . . . . GEI              // general
  //            x  . . . . . . PEIE             // periferal
  //             x . . . . . . T0IE             // timer0

  return;
  }


// ===================================================================
#if defined(LCD_LINESIZE)       /* --- LCD code --- */
// ===================================================================

// ------------------------------------------------------------
// Clear screen LCD
// ------------------------------------------------------------

#pragma update_RP 0

static void   lcdClearScreen(void) {

  RS = 0;                                       // control character follows
  lcdWrtChar(0b0000.0001);                      // clear screen, cursor home
  delay100microsecond(15);                      // wait a little
  RS = 1;                                       // switch back to data chars
  }

#pragma update_RP 1


// ------------------------------------------------------------
// Set cursor position
// ------------------------------------------------------------

#pragma update_RP 0

static void   lcdSetCursor(uns8  row,
                           uns8  col) {

  col |= 0b1000.0000;                           // add offset top line
  if (row > 0)                                  // 2nd line req'd
    col |= 0b0100.0000;                         // add offset 2nd line
  RS = 0;                                       // control character follows
  lcdWrtChar(col);                              // new cursor position
  RS = 1;                                       // switch back to data chars
  }

#pragma update_RP 1


// -----------------------------------------------------
// Initialize LCD
// - global init
// - 2-line mode, no cursor
// -----------------------------------------------------

#pragma update_RP 0

static void   lcdSetup(void) {

  RS = 0;                                       // set control mode
  delay100microsecond(150);                     // 15 ms power-up delay
  lcdWrtNibble(0b0000.0011);                    // function set
  delay100microsecond(41);                      // > 4.1 milliseconds
  lcdWrtNibble(0b0000.0011);                    // function set
  delay100microsecond(1);                       // > 100 microseconds
  lcdWrtNibble(0b0000.0011);                    // function set
  delay16microsecond(3);
  lcdWrtNibble(0b0000.0010);                    // to 4-bit mode
  delay16microsecond(3);
  lcdWrtChar(0b0010.1000);                      // 4-bits, 2 lines, 5x8 font
  lcdWrtChar(0b0001.1100);                      // cursor move right
  lcdWrtChar(0b0000.1100);                      // display on;cursor,blink off
  lcdWrtChar(0b0000.0110);                      // cursor->right, no shift
  lcdClearScreen();                             // clear screen,
                                                // (incl switch to data mode)
  }

#pragma update_RP 1

// -----------------------------------------------------
// Show diagnostic byte (or 'no', or 'er')
// -----------------------------------------------------

#pragma update_RP 0

static void   lcdShowDiag(uns8  Diag) {

  lcdClearScreen();                             // clear previous contents
  lcdSetCursor(1, 0);                           // to second line
  if (Diag == 0x50) {                           // normal condition
    lcdWrtChar('n');
    lcdWrtChar('o');
    }
  else if (Diag == 0x00) {                      // 3705/antenna error
    lcdWrtChar('e');
    lcdWrtChar('r');
    }
  else                                          // other (error) condition
    lcdWrtHexChar(Diag);                        // hex representation
  }

#pragma update_RP 1


// --------------------------------------------------------------------
// Show key and related info on LCD
//
// Key is always displayed as 16 hexadecimal digits.
// When key is completely ASCII text it will be displayed in ASCII too.
// The transponder type will be indicated (RO or RW),
// or in stead the diagnostic byte could be shown.
// --------------------------------------------------------------------

#pragma update_RP 0

static void   lcdShowKey() {

  uns8     i;                                   // counter
  uns8     x;                                   // byte of key
  BOOLEAN  bAlphaNum;                           //

  lcdClearScreen();                             // (and cursor to 0,0)

  bAlphaNum = TRUE;                             // assume alphanumeric
  for (i=0; i<8; i++) {                         // key left to right
    x = ScioBuf.f.Key[7-i];                     // LSB first
    lcdWrtHexChar(x);                           // 2 hex digits
    if (x < 32  ||  x >= 127)                   // outside ASCII range
      bAlphaNum = FALSE;                        // not ASCII
    }

  lcdSetCursor(1, 0);                           // to second line
  if (ScioBuf.f.Diag == 0x50) {                 // normal diagnose byte
    lcdWrtChar('R');
    if (ScioBuf.f.Start == START_RW)            // RW transponder
      lcdWrtChar('W');                          // read/write
    else
      lcdWrtChar('O');                          // read-only
    }
  else                                          // not RO or RW start byte
    lcdWrtHexChar(ScioBuf.f.Diag);              // show as 2 hex digits

  if (bAlphaNum) {                              // key is all ASCII
    lcdWrtChar(' ');
    lcdWrtChar(' ');
    for (i = 0; i < 8; i++)
      lcdWrtChar(ScioBuf.f.Key[7 - i]);         // key in ASCII
    }
  }

#pragma update_RP 1


// ----------------------------------------------------------
// Send byte to the LCD as 2 nibbles (most significant first)
// ----------------------------------------------------------

#pragma update_RP 0

static void   lcdWrtChar(uns8  c) {

  lcdWrtNibble(swap(c));                        // write high nibble
  lcdWrtNibble(c);                              // write low nibble
  delay16microsecond(3);                        // > 37 s
  }

#pragma update_RP 1


// ----------------------------------------------------------
// Send byte to the LCD as 2 hex chars (most significant first)
// ----------------------------------------------------------

#pragma update_RP 0

static  void  lcdWrtHexChar(uns8  c) {

  lcdWrtHexNibble(swap(c));                     // write high nibble
  lcdWrtHexNibble(c);                           // write low nibble
  delay16microsecond(3);                        // > 37 s
  }

#pragma update_RP 1


// -----------------------------------------------------
// Write low nibble of byte to LCD (as hex digit)
// -----------------------------------------------------

#pragma update_RP 0

static  void  lcdWrtHexNibble(uns8  c) {        // byte

  c = '0' + (c & 0x0F);                         // nibble hex digit 0..9
  if (c > '9')                                  // not in 0..9 range
    c += 'A' - '0' - 10;                        // correction for A..F
  lcdWrtChar(c);                                // write the char
  }

#pragma update_RP 1


// -----------------------------------------------------
// Write nibble (lower order 4 bits of byte) to LCD.
// Nibble is sent to PORTC.0..3.
// RS not touched: function can be used for commands and data
// Required wait after write expected to be done by caller,
// in 4-bit mode only required after second nibble.
// -----------------------------------------------------

#pragma update_RP 0

static void   lcdWrtNibble(uns8  lownibble) {

  PORTC &= 0xF0;                                // leave high nibble asis
  PORTC |= (lownibble & 0x0F);                  // replace low nibble
  EN = 1;                                       // EN on
  nop();                                        // wait (> 400 ns)
  EN = 0;                                       // EN off
  }

#pragma update_RP 1

// =================================================================
#endif             /* end of LCD code */
// =================================================================


