
#define  VERSIONMAJOR      (VERSION / 100)
#define  VERSIONMINOR     ((VERSION / 100) % 10)
#define  VERSIONSUFFIX     (VERSION % 10)

typedef bit BOOL, BOOLEAN;
#define  TRUE   1
#define  FALSE  0

#define INTERRUPT_VECTOR 0x04                   // offset of service routine

// PIC I/O pin usage

#pragma  bit  READY_LED    @ PORTA.0            // on: valid frame received
#pragma  bit  ERROR_LED    @ PORTA.1            // on: frame in error
//                           PORTA.2            //  (not used)
//                           PORTA.3            // /MCLR
//                           PORTA.4            // )
//                           PORTA.5            // )  resonator
//                           PORTA.6            //  )
//                           PORTA.7            //  ) (do not exist)

//                           PORTB.0            // )
//                           PORTB.1            // ) (do not exist)
//                           PORTB.2            // )
//                           PORTB.3            // )
#pragma  bit  SCIO        @  PORTB.4            // data from TMS3705
//                           PORTB.5            // UART RX
#pragma  bit  CTSinv      @  PORTB.6            // CTS signal to DTE (PC)
//                           PORTB.7            // UART TX


//                           PORTC.0            // )
//                           PORTC.1            // ) LCD data nibble
//                           PORTC.2            // )
//                           PORTC.3            // )
#pragma  bit  RS          @  PORTC.4            // LCD control
#pragma  bit  TXCT        @  PORTC.5            // control + data -> TMS3705
//                           PORTC.6            // not used
#pragma  bit  EN          @  PORTC.7            // LCD control


// basic operational constants

#define  FOSC               16000000            // oscillator freq (Hz)

#if defined(__LOCONET__)
  #define  BPSRATE           19200              // LocoBuffer bit speed
#else
  #define  BPSRATE           9600               // RS232 bit speed (TI-S2000)
#endif

#define  BRGLOADCOUNT       ((FOSC/16/BPSRATE*10-5)/10)  // SPBRG
                                                // (with BRGH=1)


#define  XMTBUFSIZE          32                 // transmit buffer size
#define  RCVBUFSIZE          64                 // receive buffer size
                                                // note: - must be power of 2
                                                //       - maximum 64
#define  DELTA               17                 // spare room in rcv buffer
                                                // for FiFo / CTS flow control
#define  CMDBUFSIZE          21                 // buffer size host commands
                                                // Max frame is for writing new
                                                // key to transponder.
                                                // (S2000 has max frame 41).

// following macro assumes the oscillator frequency is 16 MHz!
// (1 instruction cycle is 0.25 micro seconds)

#define  delay1mus(count)  {uns8 x=count; do { nop(); } while (--x > 0); }
                                                // (4 instructions for 1 s)

enum  _phase3705 {PHASE_CHARGE,                  // burst charge
                  PHASE_READ,                    // reading 3705
                  PHASE_IDLE};                   // 3705 idle

// Timeout expressed in milliseconds after start of poll cycle.

#define  TIMEOUT_CHARGE      50                 // end of charge phase
#define  TIMEOUT_READ        70                 // end of read phase
#define  TIMEOUT_IDLE        120                // end of idle phase

// Layout of data as expected from TMS3705A chip.
// During reading of the TMS3705A handled as byte array,
// for other purposes handled as structure.

typedef  struct {                               // 3705 input buffer structure
               uns8   Diag;                     // diagnose byte
               uns8   Start;                    // start byte
               uns8   Key[8];                   // tag key (LSB first)
               uns16  BCC;                      // DBCC (LSB first)
               uns8   Stop;                     // stop byte
               uns16  End;                      // end word (LSB first)
               } SCIOFRAME;                     // SCIO frame layout


// Layout of command expected from computer program.
// This is the only layout used by this reader (a subset of S2000).

typedef  struct {                               // host cmd structure
               uns8   cmd;                      // command byte
               uns8   pl;                       // packet length
               uns8   pb1;                      // power burst 1
               uns8   pb2;                      // power burst 2
               uns8   dl;                       // data length
               uns8   kwd;                      // write keyword
               uns8   pwd;                      // write password
               uns8   key[8];                   // tag key (LSB first)
               uns16  crc;                      // crc
               uns16  frm;                      // write frame
               uns8   bcc;                      // end
               } NEWKEYCOMMAND;                 // layout host frame write key

#define  CCITTPOLY              0x1021          // CRC calculation polynoom

#define  SOH                    0x01            // Start of header
#define  START_RO               0x7E            // Startbyte RO transponder
#define  START_RW               0xFE            // Startbyte R/W transponder

#define  SCIO_STARTBIT_LEVEL    1               // high (+5V)
#define  SCIO_STOPBIT_LEVEL     0               // low (gnd)


                                                // bit timing transp. writes
#define tonH    (1000 / 100)                    // high bit on period
#define tonL    (1700 / 100)                    // low bit on period
#define toffH   (1000 / 100)                    // high bit off period
#define toffL   (300  / 100)                    // low bit off period


// PC-to-Reader protocol: flags in the command byte

#define  COMMAND_SINGLE         0x00            // single command
#define  COMMAND_NORMAL_MODE    0x01            // read contin. Normal mode
#define  COMMAND_LINE_MODE      0x02            // read contin. Line mode
#define  COMMAND_VERSION        0x03            // query firmware version
#define  COMMAND_FBCC_CALC      0x04            // calculate FBCC of MPT mode
#define  COMMAND_POWER_BURST1   0x08            // power burst 1 byte present
#define  COMMAND_PAUSE          0x10            // pause spec. present
#define  COMMAND_POWER_BURST2   0x20            // power burst 2 byte present
#define  COMMAND_DATA           0x40            // data present
#define  COMMAND_EXPANSION      0x80            // command field follows

#define  COMMAND2_WRITE_TIMING  0x01            // special timing byte present
#define  COMMAND2_WIRELESS_SYNC 0x02            // wireless sync used
#define  COMMAND2_DBCC_CALC     0x04            // calculate DBCC

// Reader-to-PC protocol: flags in the status-byte

#define  STATUS_TYPE_RO         0b0000.0000     // RO transponder
#define  STATUS_TYPE_RW         0b0000.0001     // R/W transponder
#define  STATUS_TYPE_MPT        0b0000.0010     // MPT/SAMPT transponder
#define  STATUS_TYPE_OTHER      0b0000.0011     // other (eg Firmware report)
#define  STATUS_START_BYTE      0b0000.0100     // start byte was detected
#define  STATUS_DBCC_OK         0b0000.1000     // DBCC is OK
#define  STATUS_FBCC_OK         0b0001.0000     // FBCC is OK
#define  STATUS_VERSION         0b0010.0000     // Firmware version follows


// report types

enum _type  {RPT_STS, RPT_FULL, RPT_SW, RPT_NR};


// =======================================================================
// Charge/Read phase timing with Timer0.
// This is a general purpose calculation!
// Derive constants for TMR0 interrupt, based on conditions of
// PIC oscillator speed.
// Timer0 will be set to interrupt every 1 milliseconds.
// A variable 'Tmr0Counter' is updated in the Interrupt Service Routine
// and used in the mainline to determine charge/read/idle phase.
// =======================================================================

#define  CYCLES_PER_MS  (FOSC / 4 / 1000 / 2)           // TMR0 cycles

#if CYCLES_PER_MS < 100                                 // too fast!
  #error  Oscillator frequency too slow!
  #define DELAY_TMR0_1MS_PRE   0                        // } suppress
  #define DELAY_TMR0_1MS_LOAD_VALUE   0                 // } 'undefined'
#elif CYCLES_PER_MS < 256
  #define DELAY_TMR0_1MS_PRE   0                        // 1:2
  #define DELAY_TMR0_1MS_LOAD_VALUE  (256 - CYCLES_PER_MS)
#elif CYCLES_PER_MS < 512
  #define DELAY_TMR0_1MS_PRE   1                        // 1:4
  #define DELAY_TMR0_1MS_LOAD_VALUE  (256 - CYCLES_PER_MS / 2)
#elif CYCLES_PER_MS < 1024
  #define DELAY_TMR0_1MS_PRE   2                        // 1:8
  #define DELAY_TMR0_1MS_LOAD_VALUE  (256 - CYCLES_PER_MS / 4)
#elif CYCLES_PER_MS < 2048
  #define DELAY_TMR0_1MS_PRE   3                        // 1:16
  #define DELAY_TMR0_1MS_LOAD_VALUE  (256 - CYCLES_PER_MS / 8)
#elif CYCLES_PER_MS < 4096
  #define DELAY_TMR0_1MS_PRE   4                        // 1:32
  #define DELAY_TMR0_1MS_LOAD_VALUE  (256 - CYCLES_PER_MS / 16)
#elif CYCLES_PER_MS < 8192
  #define DELAY_TMR0_1MS_PRE   5                        // 1:64
  #define DELAY_TMR0_1MS_LOAD_VALUE  (256 - CYCLES_PER_MS / 32)
#elif CYCLES_PER_MS < 16384
  #define DELAY_TMR0_1MS_PRE   6                        // 1:126
  #define DELAY_TMR0_1MS_LOAD_VALUE  (256 - CYCLES_PER_MS / 64)
#elif CYCLES_PER_MS < 32768
  #define DELAY_TMR0_1MS_PRE   7                        // 1:256
  #define DELAY_TMR0_1MS_LOAD_VALUE  (256 - CYCLES_PER_MS / 128)
#else
  #error Probably specification error, CYCLES_PER_MS too high
#endif



// ===============================================
// Function prototypes
// ===============================================
static  void   calcCRC(void);
static  void   delay16microsecond(uns8);
static  void   delay100microsecond(uns8);
static  void   hostExecCommand(void);
static  uns8   hostSendByte(uns8);
static  void   hostSendReport(uns8);
static  uns8   receiveByte(void);
static  void   transferNewKey(uns8 *);
static  void   transferMCR(uns8);
static  void   picSetup(void);

#if defined(LCD_LINESIZE)                       // only with LCD display
  static  void   lcdByteToHex(uns8, uns8 *);
  static  void   lcdClearScreen(void);
  static  void   lcdSetCursor(uns8, uns8);
  static  void   lcdSetup(void);
  static  void   lcdShowDiag(uns8);
  static  void   lcdShowKey(void);
  static  void   lcdWrtChar(uns8);
  static  void   lcdWrtHexChar(uns8);
  static  void   lcdWrtHexNibble(uns8);
  static  void   lcdWrtNibble(uns8);
#endif


// -----------------------------------------------

